from peptacular.sequence import parse_modified_sequence, create_modified_sequence, strip_modifications, \
    get_semi_sequences, get_non_enzymatic_sequences, convert_to_ms2_pip_style, \
    get_left_semi_sequences, get_right_semi_sequences, add_static_mods, add_variable_mods, sequence_generator, \
    _sequence_generator_back, _sequence_generator_front, get_fragment_sequences, get_internal_fragment_sequences

import unittest


class TestPeptide(unittest.TestCase):

    def test_read_modified_peptide(self):
        # Test the case where there are no modifications in the peptide sequence
        peptide_sequence = "ACDEFG"
        expected_output = {}
        self.assertEqual(parse_modified_sequence(peptide_sequence), expected_output)

        # Test the case where there is a single modification in the peptide sequence
        peptide_sequence = "A(2)CDEFG"
        expected_output = {0: '2'}
        self.assertEqual(parse_modified_sequence(peptide_sequence), expected_output)

        # Test the case where there are multiple modifications in the peptide sequence
        peptide_sequence = "A(2)C(3)DE(1)FG"
        expected_output = {0: '2', 1: '3', 3: '1'}
        self.assertEqual(parse_modified_sequence(peptide_sequence), expected_output)

        # Test the case where the peptide sequence contains incorrect modification notation
        peptide_sequence = "A(2)CDEFG("
        with self.assertRaises(ValueError):
            parse_modified_sequence(peptide_sequence)

    def test_convert_to_ms2_pip_style(self):
        # Test the case where there are no modifications in the peptide sequence
        peptide_sequence = "ACDEFG"
        expected_output = ''
        self.assertEqual(convert_to_ms2_pip_style(peptide_sequence), expected_output)

        # Test the case where there is a single modification in the peptide sequence
        peptide_sequence = "A(Acetyl)CDEFG"
        expected_output = '1|Acetyl'
        self.assertEqual(convert_to_ms2_pip_style(peptide_sequence), expected_output)

        # Test the case where there is a single N-Term mod
        peptide_sequence = "(Acetyl)CDEFG"
        expected_output = '0|Acetyl'
        self.assertEqual(convert_to_ms2_pip_style(peptide_sequence), expected_output)

        # Test the case where there is a single C-Term mod
        peptide_sequence = "CDEFG(Acetyl)"
        expected_output = '-1|Acetyl'
        self.assertEqual(convert_to_ms2_pip_style(peptide_sequence), expected_output)

        # Test the case where there are multiple modifications in the peptide sequence
        peptide_sequence = "A(2)C(Acetyl)DE(1)FG"
        expected_output = '1|2|2|Acetyl|4|1'
        self.assertEqual(convert_to_ms2_pip_style(peptide_sequence), expected_output)

        # Test the case where the peptide sequence contains incorrect modification notation
        peptide_sequence = "A(2)CDEFG("
        with self.assertRaises(ValueError):
            convert_to_ms2_pip_style(peptide_sequence)

    def test_create_modified_peptide(self):
        # Test the case where there are no modifications in the peptide sequence
        unmodified_sequence = "ACDEFG"
        modifications = {}
        expected_output = "ACDEFG"
        self.assertEqual(create_modified_sequence(unmodified_sequence, modifications), expected_output)

        # Test the case where there is a single modification in the peptide sequence
        unmodified_sequence = "ACDEFG"
        modifications = {0: 2}
        expected_output = "A(2)CDEFG"
        self.assertEqual(create_modified_sequence(unmodified_sequence, modifications), expected_output)

        # Test the case where there is a single N-Term mod
        unmodified_sequence = "ACDEFG"
        modifications = {-1: 2}
        expected_output = "(2)ACDEFG"
        self.assertEqual(create_modified_sequence(unmodified_sequence, modifications), expected_output)

        # Test the case where there is a single C-Term mod
        unmodified_sequence = "ACDEFG"
        modifications = {5: 2}
        expected_output = "ACDEFG(2)"
        self.assertEqual(create_modified_sequence(unmodified_sequence, modifications), expected_output)

        # Test the case where there are multiple modifications in the peptide sequence
        unmodified_sequence = "ACDEFG"
        modifications = {0: 2, 1: 3, 3: 1}
        expected_output = "A(2)C(3)DE(1)FG"
        self.assertEqual(create_modified_sequence(unmodified_sequence, modifications), expected_output)

        # Test the case where the index of a modification is invalid for the given peptide sequence
        unmodified_sequence = "ACDEFG"
        modifications = {10: 2}
        with self.assertRaises(ValueError):
            create_modified_sequence(unmodified_sequence, modifications)

    def test_get_unmodified_peptide(self):
        self.assertEqual(strip_modifications('ACDEFG'), 'ACDEFG')
        self.assertEqual(strip_modifications('A(2)CDE(2)FG'), 'ACDEFG')
        self.assertEqual(strip_modifications('A(2)CDE(Acetyl)FG'), 'ACDEFG')

    def test_get_semi_sequences(self):
        assert set(get_semi_sequences(sequence='PEPTIDE', min_len=3, max_len=5)) == \
               {'TIDE', 'PTIDE', 'PEPT', 'PEPTI', 'IDE', 'PEP'}
        assert set(get_semi_sequences(sequence='PEPTIDE', min_len=2, max_len=5)) == \
               {'TIDE', 'PTIDE', 'PEPT', 'PEPTI', 'IDE', 'PEP', 'PE', 'DE'}
        assert set(get_semi_sequences(sequence='PEPTIDE', min_len=2, max_len=100)) == \
               {'TIDE', 'PTIDE', 'PEPT', 'PEPTI', 'IDE', 'PEP', 'PE', 'DE', 'PEPTID', 'EPTIDE'}
        assert set(get_semi_sequences(sequence='PEPTIDE', min_len=None, max_len=None)) == \
               {'P', 'E', 'TIDE', 'PTIDE', 'PEPT', 'PEPTI', 'IDE', 'PEP', 'PE', 'DE', 'PEPTID', 'EPTIDE'}

    def test_get_left_semi_sequences(self):
        self.assertEqual(set(get_left_semi_sequences('PEPTIDE', None, None)),
                         {'P', 'PE', 'PEP', 'PEPT', 'PEPTI', 'PEPTID'})
        self.assertEqual(set(get_left_semi_sequences('PEPTIDE', 3, None)), {'PEP', 'PEPT', 'PEPTI', 'PEPTID'})
        self.assertEqual(set(get_left_semi_sequences('PEPTIDE', None, 5)), {'P', 'PE', 'PEP', 'PEPT', 'PEPTI'})


        assert set(get_left_semi_sequences(sequence='PEPTIDE', min_len=None, max_len=None)) == \
               {'P', 'PE', 'PEP', 'PEPT', 'PEPTI', 'PEPTID'}
        assert set(get_left_semi_sequences(sequence='PEPTIDE', min_len=3, max_len=None)) == \
               {'PEP', 'PEPT', 'PEPTI', 'PEPTID'}
        assert set(get_left_semi_sequences(sequence='PEPTIDE', min_len=None, max_len=5)) == \
               {'P', 'PE', 'PEP', 'PEPT', 'PEPTI'}
        assert set(get_left_semi_sequences(sequence='PEPTIDE', min_len=3, max_len=4)) == \
               {'PEP', 'PEPT'}

    def test_get_right_semi_sequences(self):
        self.assertEqual(set(get_right_semi_sequences('PEPTIDE', None, None)),
                         {'EPTIDE', 'PTIDE', 'TIDE', 'IDE', 'DE', 'E'})
        self.assertEqual(set(get_right_semi_sequences('PEPTIDE', 3, None)), {'EPTIDE', 'PTIDE', 'TIDE', 'IDE'})
        self.assertEqual(set(get_right_semi_sequences('PEPTIDE', None, 5)), {'PTIDE', 'TIDE', 'IDE', 'DE', 'E'})
        assert set(get_right_semi_sequences(sequence='PEPTIDE', min_len=None, max_len=None)) == \
               {'EPTIDE', 'PTIDE', 'TIDE', 'IDE', 'DE', 'E'}
        assert set(get_right_semi_sequences(sequence='PEPTIDE', min_len=3, max_len=None)) == \
               {'EPTIDE', 'PTIDE', 'TIDE', 'IDE'}
        assert set(get_right_semi_sequences(sequence='PEPTIDE', min_len=None, max_len=5)) == \
               {'PTIDE', 'TIDE', 'IDE', 'DE', 'E'}
        assert set(get_right_semi_sequences(sequence='PEPTIDE', min_len=3, max_len=4)) == \
               {'TIDE', 'IDE'}

    def test_get_all_non_enzymatic_peptides(self):
        self.assertEqual(set(get_non_enzymatic_sequences('PEPT')),
                         {'P', 'E', 'P', 'T', 'PE', 'EP', 'PT', 'PEP', 'EPT', 'PEPT'})
        self.assertEqual(set(get_non_enzymatic_sequences('PEPT', min_len=1, max_len=2)),
                         {'P', 'E', 'P', 'T', 'PE', 'EP', 'PT'})
        self.assertEqual(set(get_non_enzymatic_sequences('PEPT', min_len=2, max_len=4)),
                         {'PE', 'EP', 'PT', 'PEP', 'EPT', 'PEPT'})

    def test_add_static_mod(self):
        self.assertEqual(add_static_mods('PEPCTIDE', {'C': 57.021464}), 'PEPC(57.021464)TIDE')
        self.assertEqual(add_static_mods('PEPC(57.021464)TIDE', {'C': 57.021464}), 'PEPC(57.021464)TIDE')
        self.assertEqual(add_static_mods('CPEPTIDEC', {'C': 57.021464}), 'C(57.021464)PEPTIDEC(57.021464)')
        self.assertEqual(add_static_mods('P(1)EPCTIDE(1)', {'C': 57.021464}), 'P(1)EPC(57.021464)TIDE(1)')

    def test_add_variable_mod(self):
        self.assertEqual(set(add_variable_mods('PEPCTIDCE', {'C': 57.021464}, 2)), {'PEPC(57.021464)TIDCE',
                                                                                    'PEPC(57.021464)TIDC(57.021464)E',
                                                                                    'PEPCTIDC(57.021464)E',
                                                                                    'PEPCTIDCE'})

        self.assertEqual(set(add_variable_mods('P(1)EPC(20)TIDCE', {'C': 57.021464}, 2)), {'P(1)EPC(20)TIDCE',
                                                                                        'P(1)EPC(20)TIDC(57.021464)E'})

    def test_peptide_generator(self):
        # Test forward generator
        generator = sequence_generator("A(P)C(P)G", True)
        self.assertEqual(next(generator), "A(P)C(P)G")
        self.assertEqual(next(generator), "C(P)G")
        self.assertEqual(next(generator), "G")
        with self.assertRaises(StopIteration):
            next(generator)

        # Test backward generator
        generator = sequence_generator("A(P)C(P)G", False)
        self.assertEqual(next(generator), "A(P)C(P)G")
        self.assertEqual(next(generator), "A(P)C(P)")
        self.assertEqual(next(generator), "A(P)")
        with self.assertRaises(StopIteration):
            next(generator)

    def test_peptide_generator_back(self):
        generator = _sequence_generator_back("A(P)C(P)G")
        self.assertEqual(next(generator), "A(P)C(P)G")
        self.assertEqual(next(generator), "A(P)C(P)")
        self.assertEqual(next(generator), "A(P)")
        with self.assertRaises(StopIteration):
            next(generator)

    def test_peptide_generator_front(self):
        generator = _sequence_generator_front("(P)A(P)C(P)G")
        self.assertEqual(next(generator), "(P)A(P)C(P)G")
        self.assertEqual(next(generator), "C(P)G")
        self.assertEqual(next(generator), "G")
        with self.assertRaises(StopIteration):
            next(generator)

    def test_get_fragment_sequences(self):
        sequence = "PEPTIDE"

        for ion_type in 'abc':
            fragments = get_fragment_sequences(sequence, ion_type)
            self.assertEqual(fragments, ['P', 'PE', 'PEP', 'PEPT', 'PEPTI', 'PEPTID', 'PEPTIDE'][::-1])

        for ion_type in 'xyz':
            fragments = get_fragment_sequences(sequence, ion_type)
            self.assertEqual(fragments, ['E', 'DE', 'IDE', 'TIDE', 'PTIDE', 'EPTIDE', 'PEPTIDE'][::-1])

    def test_get_internal_fragment_sequences(self):
        sequence = "PEPTIDE"

        for ion_type in 'xyz':
            fragments = get_internal_fragment_sequences(sequence, ion_type)
            self.assertEqual(fragments, ['P', 'PE', 'PEP', 'PEPT', 'PEPTI', 'PEPTID', 'PEPTIDE'][::-1])

        for ion_type in 'abc':
            fragments = get_internal_fragment_sequences(sequence, ion_type)
            self.assertEqual(fragments, ['E', 'DE', 'IDE', 'TIDE', 'PTIDE', 'EPTIDE', 'PEPTIDE'][::-1])

    def test_get_fragment_sequences_modified(self):
        sequence = "(-10)PEP(2)TIDE(100)"

        for ion_type in 'abc':
            fragments = get_fragment_sequences(sequence, ion_type)
            self.assertEqual(fragments, ['(-10)P', '(-10)PE', '(-10)PEP(2)', '(-10)PEP(2)T', '(-10)PEP(2)TI', '(-10)PEP(2)TID', '(-10)PEP(2)TIDE(100)'][::-1])

        for ion_type in 'xyz':
            fragments = get_fragment_sequences(sequence, ion_type)
            self.assertEqual(fragments, ['E(100)', 'DE(100)', 'IDE(100)', 'TIDE(100)', 'P(2)TIDE(100)', 'EP(2)TIDE(100)', '(-10)PEP(2)TIDE(100)'][::-1])

    def test_get_internal_fragment_sequences_modified(self):
        sequence = "(-10)PEP(2)TIDE(100)"

        for ion_type in 'xyz':
            fragments = get_internal_fragment_sequences(sequence, ion_type)
            self.assertEqual(fragments, ['(-10)P', '(-10)PE', '(-10)PEP(2)', '(-10)PEP(2)T', '(-10)PEP(2)TI', '(-10)PEP(2)TID', '(-10)PEP(2)TIDE(100)'][::-1])

        for ion_type in 'abc':
            fragments = get_internal_fragment_sequences(sequence, ion_type)
            self.assertEqual(fragments, ['E(100)', 'DE(100)', 'IDE(100)', 'TIDE(100)', 'P(2)TIDE(100)', 'EP(2)TIDE(100)', '(-10)PEP(2)TIDE(100)'][::-1])


if __name__ == "__main__":
    unittest.main()



from functools import wraps
from typing import Tuple, List, Optional
from itertools import groupby

# TODO: Remove wrapper function? Its confusing and hurts readability


def span_processing(func):
    """
    A decorator to enforce constraints on the span and min_len and max_len arguments of decorated functions.
    The span should be a tuple of length 3 and min_len and max_len should be None or an integer.
    """

    @wraps(func)
    def wrapper(span: Tuple[int, int, int], min_len: Optional[int] = None, max_len: Optional[int] = None):
        assert isinstance(span, Tuple) and len(span) == 3, 'span should be a tuple of length 3.'
        assert all(isinstance(arg, (int, type(None))) for arg in
                   (min_len, max_len)), 'min_len and max_len should be None or an integer.'
        _validate_span(span)

        if min_len is None:
            min_len = 1
        if max_len is None:
            max_len = span[1] - span[0]

        return func(span, min_len, max_len)

    return wrapper


def _validate_span(span: Tuple[int, int, int]) -> None:
    """
    This function checks if a given span is valid.

    Args:
        span (Tuple[int, int, int]): A tuple representing the span, structured as (start, end, value).

    Raises:
        ValueError: If the span is not valid.

    """
    start, end, value = span
    if start < 0:
        raise ValueError(f'Start of span should be non-negative, got {start}.')
    if end < 0:
        raise ValueError(f'End of span should be non-negative, got {end}.')
    if start > end:
        raise ValueError(f'Start of span {start} should be less than or equal to end of span {end}.')


@span_processing
def build_non_enzymatic_spans(span: Tuple[int, int, int], min_len: int, max_len: int) -> List[Tuple[int, int, int]]:
    """
    This function generates and returns all possible sub-spans of the given span. These sub-spans have lengths ranging
    from `min_len` to `max_len`. The sub-spans are "non-enzymatic", meaning they are direct subsets of the given span
    and not created through any enzymatic action or process.

    Args:
        span (Tuple[int, int, int]): A tuple representing the original span, structured as (start, end, value).
        min_len (int, optional): The minimum length of sub-spans to be generated. Defaults to 1.
        max_len (int, optional): The maximum length of sub-spans to be generated. If not provided, it defaults to the
                                 length of the input span.

    Returns:
        List[Tuple[int, int, int]]: A list of all possible sub-spans as tuples, each structured as (start, end, value).
    """
    start, end, value = span
    return [(i, j, 0) for i in range(start, end + 1) for j in range(i + min_len, min(end + 1, i + max_len + 1))]


@span_processing
def build_left_semi_spans(span: Tuple[int, int, int], min_len: int, max_len: int) -> List[Tuple[int, int, int]]:
    """
    This function generates and returns all possible sub-spans of the given span starting from the left.
    These sub-spans have lengths ranging from `min_len` to `max_len`.

    Args:
        span (Tuple[int, int, int]): A tuple representing the original span, structured as (start, end, value).
        min_len (int, optional): The minimum length of sub-spans to be generated. Defaults to 1.
        max_len (int, optional): The maximum length of sub-spans to be generated. If not provided, it defaults to the
                                 length of the input span.

    Returns:
        List[Tuple[int, int, int]]: A list of all possible left sub-spans as tuples, each structured as
                                    (start, end, value).
    """
    if min_len > max_len:
        return []

    start, end, value = span
    new_end = min(start + max_len, end - 1)
    return [(start, i, value) for i in range(new_end, start - 1, -1) if i - start >= min_len]


@span_processing
def build_right_semi_spans(span: Tuple[int, int, int], min_len: int, max_len: int) -> List[Tuple[int, int, int]]:
    """
    This function generates and returns all possible sub-spans of the given span starting from the right.
    These sub-spans have lengths ranging from `min_len` to `max_len`.

    Args:
        span (Tuple[int, int, int]): A tuple representing the original span, structured as (start, end, value).
        min_len (int, optional): The minimum length of sub-spans to be generated. Defaults to 1.
        max_len (int, optional): The maximum length of sub-spans to be generated. If not provided, it defaults to the
                                 length of the input span.

    Returns:
        List[Tuple[int, int, int]]: A list of all possible right sub-spans as tuples, each structured as
                                    (start, end, value).
    """

    if min_len > max_len:
        return []

    start, end, value = span
    new_start = max(start + 1, end - max_len)
    return [(i, end, value) for i in range(new_start, end + 1) if end - i >= min_len]


def span_to_sequence(sequence: str, span: Tuple[int, int, int]) -> str:
    """
    This function takes a sequence and a span as input, then returns the subsequence of the input sequence
    that corresponds to the provided span.

    Args:
        sequence (str): The original sequence from which a subsequence will be extracted.
        span (Tuple[int, int, int]): A tuple representing the span of the subsequence to be extracted, structured as
                                     (start, end, value).

    Returns:
        str: The subsequence of the input sequence as defined by the start and end of the provided span.
    """

    _validate_span(span)
    return sequence[span[0]:span[1]]


def get_enzymatic_spans(max_index: int, enzyme_sites: List[int], missed_cleavages: int,
                        min_len: int = None, max_len: int = None) -> List[Tuple[int, int, int]]:
    """
    Computes the enzymatic spans for the given enzymatic sites and the number of missed
    cleavages.

    Parameters:
        max_index (int): The max index of the span.
        enzyme_sites (List[int]): The list of indices in the protein sequence that represent enzyme sites.
        missed_cleavages (int): The number of allowed missed cleavages.
        min_len (int): The minimum length of an enzymatic span.
        max_len (int): The maximum length of an enzymatic span.

    Returns:
        List[Tuple[int, int, int]]: A list of tuples, where each tuple represents a span and contains three integers:
                                    the start index, the end index, and the number of missed cleavages.
    """

    if min_len is None:
        min_len = 1
    if max_len is None:
        max_len = max_index

    enzyme_sites = [0] + enzyme_sites + [max_index]

    spans = []
    for i, start_site in enumerate(enzyme_sites):
        spans.extend([(start_site, end_site, j) for j, end_site in enumerate(enzyme_sites[i+1:i+missed_cleavages+2])])

    # Filter spans based on length
    spans = [span for span in spans if min_len <= span[1] - span[0] <= max_len]

    return spans


def _get_all_left_semi_spans(spans: List[Tuple[int, int, int]], min_len: int, max_len: int) -> \
        List[Tuple[int, int, int]]:
    """
    Get all left semi-spans from the given list of spans that have a length within the specified range.

    :param spans: A list of tuples, where each tuple represents a span with three integers (start, end, value).
    :param min_len: The minimum length of the left semi-spans.
    :param max_len: The maximum length of the left semi-spans.
    :return: A list of tuples representing all left semi-spans that are within the specified length range.

    This function groups the spans by the start position, for each group, it checks every span to see if its length
    is at least the minimum length. If it is, it calculates the new maximum length, which is the smallest of the max_len
    and the span's length. Then it adds the left semi-spans to the list of semi-spans.
    """
    semi_spans = []
    spans = sorted(spans, key=lambda x: (x[0], -x[2]))
    for _, group in groupby(spans, key=lambda x: x[0]):
        group = list(group)
        for i, span in enumerate(group):
            span_len = span[1] - span[0]

            if span_len <= min_len:
                break

            new_max_len = min(max_len, span_len - 1)
            if i == len(group) - 1:
                semi_spans.extend(build_left_semi_spans(span, min_len, new_max_len))
            else:
                next_span = group[i + 1]
                next_span_len = next_span[1] - next_span[0]
                new_min = max(min_len, next_span_len + 1)
                semi_spans.extend(build_left_semi_spans(span, new_min, new_max_len))

    return semi_spans


def _get_all_right_semi_spans(spans: List[Tuple[int, int, int]], min_len: int, max_len: int) -> \
        List[Tuple[int, int, int]]:
    """
    Get all right semi-spans from the given list of spans that have a length within the specified range.

    :param spans: A list of tuples, where each tuple represents a span with three integers (start, end, value).
    :param min_len: The minimum length of the right semi-spans.
    :param max_len: The maximum length of the right semi-spans.
    :return: A list of tuples representing all right semi-spans that are within the specified length range.

    This function groups the spans by the start position, for each group, it checks every span to see if its length
    is at least the minimum length. If it is, it calculates the new maximum length, which is the smallest of the max_len
    and the span's length. Then it adds the right semi-spans to the list of semi-spans.
    """

    semi_spans = []
    spans = sorted(spans, key=lambda x: (x[1], -x[2]))
    for _, group in groupby(spans, key=lambda x: x[1]):
        group = list(group)
        for i, span in enumerate(group):
            span_len = span[1] - span[0]

            if span_len < min_len:
                break

            new_max_len = min(max_len, span_len - 1)
            if i == len(group) - 1:
                semi_spans.extend(build_right_semi_spans(span, min_len, new_max_len))
            else:
                next_span = group[i + 1]
                next_span_len = next_span[1] - next_span[0]
                new_min = max(min_len, next_span_len + 1)
                semi_spans.extend(build_right_semi_spans(span, new_min, new_max_len))

    return semi_spans


def get_semi_spans(spans: List[Tuple[int, int, int]], min_len: int, max_len: int) -> \
        List[Tuple[int, int, int]]:
    """
    Computes the semi spans for a given list of spans based on the given minimum and maximum length criteria.

    Parameters:
        spans (List[Tuple[int, int, int]]): The list of spans for which the semi spans will be computed.
        min_len (int): The minimum length of a semi span.
        max_len (int): The maximum length of a semi span.

    Returns:
        List[Tuple[int, int, int]]: A list of tuples, where each tuple represents a semi span and contains three
                                    integers: the start index, the end index, and the number of missed cleavages.
    """

    semi_spans = []

    spans = sorted(spans, key=lambda x: (x[0], -x[2]))
    for _, group in groupby(spans, key=lambda x: x[0]):
        semi_spans.extend(_get_all_left_semi_spans(list(group), min_len, max_len))

    spans = sorted(spans, key=lambda x: (x[1], -x[2]))
    for _, group in groupby(spans, key=lambda x: x[1]):
        semi_spans.extend(_get_all_right_semi_spans(list(group), min_len, max_len))

    return semi_spans


def build_spans(max_index: int, enzyme_sites: List[int], missed_cleavages: int, min_len: int,
                max_len: int, semi: bool) -> List[Tuple[int, int, int]]:
    """
    Builds spans for a given sequence based on the given enzyme sites, the number of missed cleavages, and the minimum
    and maximum length criteria. If semi is set to True, semi spans are also computed and added to the list of spans.

    Parameters:
        max_index (int): The max index of the span.
        enzyme_sites (List[int]): The list of indices in the sequence that represent enzyme sites.
        missed_cleavages (int): The number of allowed missed cleavages.
        min_len (int): The minimum length of a span.
        max_len (int): The maximum length of a span.
        semi (bool): Whether to compute and include semi spans.

    Returns:
        List[Tuple[int, int, int]]: A list of tuples, where each tuple represents a span and contains three integers:
                                    the start index, the end index, and the number of missed cleavages.
    """

    if min_len is None:
        min_len = 1
    if max_len is None:
        max_len = max_index

    if semi:
        spans = get_enzymatic_spans(max_index, enzyme_sites, missed_cleavages, None, None)
        semi_spans = get_semi_spans(spans, min_len, max_len)

        # filter spans by len
        spans = [span for span in spans if max_len >= span[1] - span[0] >= min_len] + semi_spans
    else:
        spans = get_enzymatic_spans(max_index, enzyme_sites, missed_cleavages, min_len, max_len)
    return spans



from typing import List

import numpy as np

from .constants import *
from peptacular.sequence import parse_modified_sequence, strip_modifications, sequence_generator, \
    identify_cleavage_sites


# TODO: Remove numpy dependency and separate index functions into separate project


def filter_by_mass(sequences: list[str], min_mass: float = None, max_mass: float = None) -> List[str]:
    """
    Filters a list of sequences by mass, returning only those that fall within the specified range.

    Args:
        sequences (list[str]): A list of sequences to be filtered.
        min_mass (float): Minimum mass of the returned sequences.
        max_mass (float): Maximum mass of the returned sequences.

    Returns:
        list[str]: A list of sequences that fall within the specified mass range.
    """

    min_mass = min_mass or 0
    max_mass = max_mass or float('inf')

    return [sequence for sequence in sequences if min_mass <= calculate_mass(sequence) <= max_mass]


def calculate_mass(sequence: str, charge=0, ion_type: str = 'y', monoisotopic=True) -> float:
    """
    Calculate the mass of a peptide sequence considering possible modifications, charge and ion type.

    Args:
        sequence (str): Peptide sequence, possibly including modifications.
        charge (int, optional): Peptide charge. Default is 0.
        ion_type (str, optional): Ion type ('a', 'b', 'c', 'x', 'y', 'z'). Default is 'y'.
        monoisotopic (bool, optional): If true, uses monoisotopic masses. Defaults to True.

    Returns:
        float: The calculated mass of the peptide sequence.
    """
    # Select mass set based on monoisotopic flag
    atomic_masses, aa_masses = (MONO_ISOTOPIC_ATOMIC_MASSES, MONO_ISOTOPIC_AA_MASSES) if monoisotopic else \
        (AVERAGE_ATOMIC_MASSES, AVERAGE_AA_MASSES)

    # Parse modifications and strip them from sequence
    mods = parse_modified_sequence(sequence)
    stripped_sequence = strip_modifications(sequence)

    # Calculate mass
    mass = sum(aa_masses[aa] for aa in stripped_sequence) + sum(float(value) for value in mods.values()) + \
        (charge * atomic_masses['PROTON'])

    # Adjust mass based on ion type
    ion_adjustments = {
        'a': -(atomic_masses['CARBON'] + atomic_masses['OXYGEN']),
        'c': atomic_masses['HYDROGEN'] * 3 + atomic_masses['NITROGEN'],
        'x': atomic_masses['CARBON'] + atomic_masses['OXYGEN'] * 2,
        'y': atomic_masses['HYDROGEN'] * 2 + atomic_masses['OXYGEN'],
        'z': atomic_masses['OXYGEN'] - atomic_masses['NITROGEN'] - atomic_masses['HYDROGEN']
    }

    return mass + ion_adjustments.get(ion_type, 0)


def calculate_mz(sequence: str, charge=0, ion_type: str = 'y', monoisotopic=True) -> float:
    """
    Calculate the m/z (mass-to-charge ratio) of a peptide sequence.

    Args:
        sequence (str): Peptide sequence which may include modifications.
        charge (int, optional): Charge of the peptide. Default is 0.
        ion_type (str, optional): Ion type, can be one of 'a', 'b', 'c', 'x', 'y', 'z'. Default is 'y'.
        monoisotopic (bool, optional): If true, uses monoisotopic masses. Defaults to True.

    Returns:
        float: The m/z of the peptide sequence. If charge is 0, returns the mass.
    """
    # Calculate mass
    mass = calculate_mass(sequence, charge, ion_type, monoisotopic)

    # Return mass if charge is zero; otherwise, return m/z
    return mass if charge == 0 else mass / charge


def fragment_sequence(sequence: str, types=('b', 'y'), max_charge=1, monoisotopic=True):
    """
    Generates fragments of a given amino acid sequence based on the specified ion types and maximum charge.
    This function is a generator that yields the m/z (mass-to-charge ratio) of each fragment.

    Args:
        sequence (str): The amino acid sequence to be fragmented.
        types (tuple, optional): A tuple containing the types of ions to be considered in the fragmentation.
            Each ion type is represented by a single character. Defaults to ('b', 'y').
        max_charge (int, optional): The maximum charge to consider for the fragments. Defaults to 1.
        monoisotopic (bool, optional): Indicates whether to calculate monoisotopic mass. Defaults to True.

    Yields:
        float: The calculated m/z for the fragment.
    """
    # Loop through the ion types and charges
    for ion_type in types:
        for charge in range(1, max_charge + 1):
            # If the ion type is in 'xyz' or 'abc', yield the series
            if ion_type in 'xyzabc':
                yield from fragment_series(sequence, ion_type=ion_type, charge=charge, monoisotopic=monoisotopic)


def fragment_series(sequence: str, ion_type='y', charge=1, monoisotopic=True):
    """
    Generates fragment series based on ion type and charge.

    Args:
        sequence (str): The amino acid sequence to be fragmented.
        ion_type (str, optional): The type of ions to be considered in the fragmentation.
            Each ion type is represented by a single character. Defaults to 'y'.
        charge (int, optional): The charge to consider for the fragments. Defaults to 1.
        monoisotopic (bool, optional): Indicates whether to calculate monoisotopic mass. Defaults to True.

    Yields:
        float: The calculated m/z for the fragment.
    """
    # Check if ion type is forward or reverse
    forward = ion_type in 'xyz'

    # Loop through the sequence to generate fragments and yield m/z
    for pep in sequence_generator(sequence, forward=forward):
        yield calculate_mz(pep, charge=charge, ion_type=ion_type, monoisotopic=monoisotopic)


def calculate_mass_array(sequence: str, monoisotopic: bool = True):
    if monoisotopic is True:
        aa_masses = MONO_ISOTOPIC_AA_MASSES
    else:
        aa_masses = AVERAGE_AA_MASSES

    sequence_mods = parse_modified_sequence(sequence)
    stripped_sequence = strip_modifications(sequence)

    mass_arr = np.zeros(len(stripped_sequence), dtype=np.float32)
    for i, aa in enumerate(stripped_sequence):

        mod_mass = 0
        if i in sequence_mods:
            mod_mass += float(sequence_mods[i])

        if i == 0 and -1 in sequence_mods:
            mod_mass += float(sequence_mods[-1])

        mass_arr[i] = np.float32(aa_masses[aa] + mod_mass)

    mass_arr = np.array(mass_arr, dtype=np.float32)

    return mass_arr


def create_ion_table(sequence: str, max_len: int = 50, ion_type: str = 'y', charge: int = 0,
                     enzyme: str = 'non-specific'):
    if ion_type == 'y':
        sequence = sequence[::-1]

    sequence_cumulative_sum = np.cumsum(calculate_mass_array(sequence))
    len_sequence = len(sequence)

    # Add padding zeros to make our life easier when computing window sums
    sequence_cumulative_sum = np.pad(sequence_cumulative_sum, (1, 0), 'constant')

    # Window start and end indices
    window_start = np.arange(len_sequence)[:, None]
    window_end = window_start + np.arange(1, max_len + 1)

    # Limit window_end indices to array bounds
    window_end = np.minimum(window_end, len_sequence)
    # Compute window sums
    ion_table = sequence_cumulative_sum[window_end] - sequence_cumulative_sum[window_start]

    # Add (HYDROGEN * 2 + OXYGEN) to the computed mass, only where window_end > window_start
    if ion_type == 'y':
        ion_table[window_end > window_start] += (MONO_ISOTOPIC_ATOMIC_MASSES['HYDROGEN'] * 2 +
                                                 MONO_ISOTOPIC_ATOMIC_MASSES['OXYGEN']) + \
                                                (MONO_ISOTOPIC_ATOMIC_MASSES['PROTON'] * charge)
    elif ion_type == 'b':
        ion_table[window_end > window_start] += (MONO_ISOTOPIC_ATOMIC_MASSES['PROTON'] * charge)
    else:
        raise ValueError("Invalid ion type. Must be either 'y' or 'b'")

    if charge != 0:
        ion_table = ion_table / charge

    # set invalid indexes to 0
    for i, j in enumerate(range(len(ion_table) - 1, len(ion_table) - min(max_len, len_sequence) - 1, -1)):
        ion_table[j, i + 1:] = 0

    if len_sequence < max_len:
        ion_table[:, len_sequence:] = 0

    if enzyme != 'non-specific':
        filter_ion_table(ion_table, sequence, enzyme)

    return ion_table


def filter_ion_table(ion_table: np.ndarray, sequence: str, enzyme: str):
    enzyme_sites = set(identify_cleavage_sites(sequence, enzyme))

    for i in range(len(ion_table)):
        if i not in enzyme_sites:
            ion_table[i, :] = 0


def get_sorted_ion_table_indexes(matrix: np.ndarray) -> np.ndarray:
    # Get 2D indexes without flattening
    # Apply mass filters if provided
    indexes_2d = np.argwhere(matrix)

    # Sort by values at these indexes
    sorted_2d_indexes = indexes_2d[np.argsort(matrix[indexes_2d[:, 0], indexes_2d[:, 1]], kind='mergesort')]
    return sorted_2d_indexes



import re
from copy import deepcopy
from typing import Dict, List, Any, Union, Generator

import regex as reg

from .constants import PROTEASES
from .spans import span_to_sequence, build_spans, build_non_enzymatic_spans, build_left_semi_spans, \
    build_right_semi_spans


# TODO: Figure out how to specify modifications: either dict[int, str] or dict[int, str]?


def _are_parentheses_balanced(text):
    """
    Check if parentheses in the given text are balanced or not.

    The function uses a stack data structure to ensure each opening parenthesis '('
    has a corresponding closing parenthesis ')' and vice versa.

    Args:
        text (str): The input string to check for balanced parentheses.

    Returns:
        bool: True if parentheses are balanced, False otherwise.
    """
    stack = []
    for i, char in enumerate(text):
        if char == '(':
            stack.append(i)
        elif char == ')':
            if not stack:
                return False
            else:
                stack.pop()
    return len(stack) == 0


def convert_to_ms2_pip_style(sequence: str) -> str:
    """
    Converts a peptide sequence to the format used by MS2PIP prediction software.

    This function parses a peptide sequence and transforms it into a string representation
    where each modification's location is paired with the type of modification, separated
    by a pipe ('|') symbol.

    Args:
        sequence (str): The peptide sequence, potentially including modifications.

    Returns:
        str: The sequence transformed into MS2PIP format.

    Example:
        >>> convert_to_ms2_pip_style('PEP(Phospho)TIDE(Phospho)K')
        '2|Phospho|6|Phospho'
    """

    # Parse modifications from the sequence
    modification_dict = parse_modified_sequence(sequence)

    ms2_pip_pairs = []
    for location, modification in modification_dict.items():
        # Adjust location index for MS2PIP
        ms2_pip_location = location + 1
        if ms2_pip_location == len(strip_modifications(sequence)):
            ms2_pip_location = -1

        # Append the new location and modification pair
        ms2_pip_pairs.append(f'{ms2_pip_location}|{modification}')

    # Combine pairs into a single string, separated by '|'
    return '|'.join(ms2_pip_pairs)


def parse_modified_sequence(sequence: str) -> Dict[int, str]:
    """
    Parses a peptide sequence with modifications and returns a dictionary where keys
    represent the position of the modified amino acid and values are the respective modifications.

    Args:
        sequence (str): The peptide sequence, potentially including modifications.

    Returns:
        Dict[int, str]: A dictionary with the modification values indexed by the position
                        of the modified amino acid.

    Raises:
        ValueError: If the sequence contains unmatched parentheses indicating incorrect
                    modification notation.

    Example:
        >>> parse_modified_sequence('PEP(Phospho)TIDE')
        {3: 'Phospho'}
    """

    # Check parentheses balance
    if not _are_parentheses_balanced(sequence):
        raise ValueError(f'Incorrect modification notation in peptide sequence : {sequence}!')

    mod_pattern = re.compile(r'\(([^)]*)\)')
    matches = mod_pattern.finditer(sequence)

    modifications = {}
    mod_offset = 0

    for match in matches:
        modification_position = match.start() - mod_offset - 1
        modification_value = match.group(1)
        modifications[modification_position] = modification_value

        # Compute offset for the next modification
        mod_offset += len(match.group())

    return modifications


def create_modified_sequence(sequence: str, modifications: Dict[int, Any]) -> str:
    """
    Generates a peptide sequence with modifications based on an unmodified sequence and a dictionary of modifications.

    The modifications are provided as a dictionary where keys correspond to the indices of the amino acids
    in the sequence to be modified, and values are the modifications to be applied. Modifications are inserted
    into the sequence in descending order of their indices to ensure their validity after each modification.

    Args:
        sequence (str): Unmodified peptide sequence.
        modifications (Dict[int, str]): Dictionary with indices of the modified amino acids and corresponding
                                        modifications.

    Returns:
        str: The peptide sequence with the specified modifications.

    Raises:
        ValueError: If an index in modifications is invalid for the sequence.

    Example:
        >>> create_modified_sequence('PEPTIDE', {3: 5.0})
        'PEP(5.0)TIDE'
    """

    modified_sequence = []
    prev_index = 0

    # Sort the modifications by index in descending order
    for i, mod in sorted(modifications.items()):

        if i < -1 or i >= len(sequence):
            raise ValueError(f'Index of modification: {i} is invalid for peptide sequence: {sequence}')

        # Insert the modification into the modified peptide sequence
        modified_sequence.append(sequence[prev_index: i + 1])
        modified_sequence.append(f"({mod})")
        prev_index = i + 1

    modified_sequence.append(sequence[prev_index:])
    return ''.join(modified_sequence)


def strip_modifications(sequence: str) -> str:
    """
    Remove any non-amino-acid characters from the given peptide sequence. This function specifically removes
    the modifications which are generally annotated within parentheses.

    Args:
        sequence (str): The peptide sequence to be processed.

    Raises:
        ValueError: If the parentheses in the given sequence are unbalanced.

    Returns:
        str: The peptide sequence without any modification annotations.
    """
    if _are_parentheses_balanced(sequence) is False:
        raise ValueError(f'Incorrect modification notation in peptide sequence : {sequence}!')

    unmodified_sequence = re.sub(r'\([^)]*\)', '', sequence)
    return unmodified_sequence


def get_left_semi_sequences(sequence: str, min_len: int = 1, max_len: int = None) -> List[str]:
    """
    Returns a set of left substrings of string `sequence` that have lengths between `min_len` and `max_len`.

    Args:
        sequence (str): The string to generate subsequences from.
        min_len (int, optional): The minimum length of the subsequence. Defaults to 1.
        max_len (int, optional): The maximum length of the subsequence. If not provided, it defaults to the length of
                                 the sequence minus one.

    Returns:
        set: A list of left subsequences of the input sequence.
    """
    spans = build_left_semi_spans((0, len(sequence), 0), min_len, max_len)
    return [span_to_sequence(sequence, span) for span in spans]


def get_right_semi_sequences(sequence: str, min_len: int = 1, max_len: int = None) -> List[str]:
    """
    Returns a set of right substrings of string `sequence` that have lengths between `min_len` and `max_len`.

    Args:
        sequence (str): The string to generate subsequences from.
        min_len (int, optional): The minimum length of the subsequence. Defaults to 1.
        max_len (int, optional): The maximum length of the subsequence. If not provided, it defaults to the length of
                                 the sequence minus one.

    Returns:
        set: A list of right subsequences of the input sequence.
    """

    spans = build_right_semi_spans((0, len(sequence), 0), min_len, max_len)
    return [span_to_sequence(sequence, span) for span in spans]


def get_semi_sequences(sequence: str, min_len: int = None, max_len: int = None) -> List[str]:
    """
    Generate a list of all semi-enzymatic amino acid sequences of a given string sequence.

    Args:
        sequence (str): A string representing the protein sequence.
        min_len (int, optional): The minimum length of the semi-enzymatic sequence. Default is None.
        max_len (int, optional): The maximum length of the semi-enzymatic sequence. Default is None.

    Returns:
        List[str]: A list of semi-enzymatic sequences within the defined length bounds.
    """

    return get_left_semi_sequences(sequence, min_len, max_len) + get_right_semi_sequences(sequence, min_len, max_len)


def get_non_enzymatic_sequences(sequence: str, min_len: int = None, max_len: int = None) -> List[str]:
    """
    Generate a list of all non-enzymatic amino acid sequences of a given string sequence.

    Args:
        sequence (str): A string representing the protein sequence.
        min_len (int, optional): The minimum length of the non-enzymatic sequence. Default is None.
        max_len (int, optional): The maximum length of the non-enzymatic sequence. Default is None.

    Returns:
        List[str]: A list of non-enzymatic sequences within the defined length bounds.
    """

    spans = build_non_enzymatic_spans((0, len(sequence), 0), min_len, max_len)
    return [span_to_sequence(sequence, span) for span in spans]


def get_enzymatic_sequences(sequence: str, enzyme_regex: str, missed_cleavages: int, semi: bool, min_len: int = None,
                            max_len: int = None) -> List[str]:
    """
    Generate a list of all enzymatic amino acid sequences of a given string sequence based on provided enzyme_regex.

    Args:
        sequence (str): A string representing the protein sequence.
        enzyme_regex (str): The enzyme regular expression for identifying cleavage sites.
        missed_cleavages (int): The maximum number of missed cleavages.
        semi (bool): Whether the sequence includes semi enzymatic peptides.
        min_len (int, optional): The minimum length of the enzymatic sequence. Default is None.
        max_len (int, optional): The maximum length of the enzymatic sequence. Default is None.

    Returns:
        List[str]: A list of enzymatic sequences within the defined length bounds and satisfying the enzyme cleavage
                   rules.
    """

    cleavage_sites = identify_cleavage_sites(sequence, enzyme_regex)
    spans = build_spans(len(sequence), cleavage_sites, missed_cleavages, min_len, max_len, semi)
    return [span_to_sequence(sequence, span) for span in spans]


def add_static_mods(sequence: str, mod_map: Dict[str, Any]):
    """
    Add static modifications to an amino acid sequence.

    Args:
        sequence (str): Original amino acid sequence.
        mod_map (Dict[str, float]): Dictionary mapping amino acids to the mass of their modifications.

    Returns:
        str: Modified amino acid sequence.
    """

    stripped_sequence = strip_modifications(sequence)
    sequence_mod_map = parse_modified_sequence(sequence)
    for aa, mass in mod_map.items():
        indexes = [i for i, x in enumerate(stripped_sequence) if x == aa]
        for index in indexes:
            sequence_mod_map[index] = mass

    return create_modified_sequence(stripped_sequence, sequence_mod_map)


def rec_mod_builder(mods: Dict[str, float], sequence: str, i: int, mod_dict: Dict[int, float], max_mod_count: int):
    """
    Recursive function to generate all possible combinations of modifications on an amino acid sequence.

    Args:
        mods (Dict[str, float]): Dictionary mapping amino acids to the mass of their modifications.
        sequence (str): amino acid sequence to modify.
        i (int): Current index on the amino acid sequence.
        mod_dict (Dict[int, float]): Dictionary mapping the indices of the amino acids to the mass of their
        modifications.
        max_mod_count (int): Maximum number of modifications allowed on the amino acid sequence.

    Yields:
        Dict[int, float]: Modified dictionary mapping the indices of the amino acid sequence to the mass of their
        modifications.
    """

    if i == len(sequence):
        yield mod_dict
        return

    original_mod_dict = deepcopy(mod_dict)
    for aa, mass in mods.items():

        if len(mod_dict) >= max_mod_count:
            break

        if sequence[i] == aa and i not in mod_dict:
            mod_dict[i] = mass
            yield from rec_mod_builder(mods, sequence, i + 1, mod_dict, max_mod_count)

    yield from rec_mod_builder(mods, sequence, i + 1, original_mod_dict, max_mod_count)


def add_variable_mods(sequence: str, mod_map: Dict[str, float], max_mods: int) -> List[str]:
    """
    Add variable modifications to an amino acid sequence.

    Args:
        sequence (str): Original amino acid sequence.
        mod_map (Dict[str, float]): Dictionary mapping amino acids to the mass of their modifications.
        max_mods (int): Maximum number of modifications allowed on the peptide.

    Returns:
        List[str]: List of all possible modified peptide sequences, each represented as a string.
    """

    original_mods = parse_modified_sequence(sequence)
    stripped_sequence = strip_modifications(sequence)
    mods = rec_mod_builder(mod_map, stripped_sequence, 0, original_mods, max_mods + len(original_mods))
    return [create_modified_sequence(stripped_sequence, mod) for mod in mods]


def sequence_generator(sequence: str, forward: bool) -> Generator[str, None, None]:
    """
    Generates amino acid sequences either from the front or the back.

    Args:
        sequence (str): The initial amino acid sequence.
        forward (bool): If True, start generating from the front; else, start from the back.

    Returns:
        generator: A generator yielding amino acid sequences.
    """
    if forward:
        return _sequence_generator_front(sequence)
    else:
        return _sequence_generator_back(sequence)


def _sequence_generator_back(sequence: str) -> Generator[str, None, None]:
    """
    Generates amino acid sequences starting from the back of the sequence.

    Args:
        sequence (str): The initial amino acid sequence.

    Yields:
        str: The sequence with the back amino acid (and modification) removed.

    Returns:
        None: If the modification starts at the beginning of the sequence.
    """
    while sequence:
        index = 0
        if sequence[-1] == ')':
            start_of_modification = sequence.rfind('(')

            if start_of_modification == 0:
                return None

            index = start_of_modification

        yield sequence
        sequence = sequence[:index - 1]


def _sequence_generator_front(sequence: str) -> Generator[str, None, None]:
    """
    Generates sub-sequences starting from the front of the sequence.

    Args:
        sequence (str): The initial sequence.

    Yields:
        str: The sequence with the front amino acid (and modification) removed.

    Returns:
        None: If the modification ends at the end of the sequence.
    """
    if sequence[0] == '(':
        yield sequence
        end_of_modification = sequence.find(')')
        sequence = sequence[end_of_modification + 2:]

        if sequence == '':
            return None

        if sequence[0] == '(':
            end_of_modification = sequence.find(')')
            sequence = sequence[end_of_modification + 1:]

    while sequence:
        index = 0
        if len(sequence) > 1 and sequence[1] == '(':
            end_of_modification = sequence.find(')')

            if end_of_modification == len(sequence):
                return None

            index = end_of_modification

        yield sequence
        sequence = sequence[index + 1:]


def reverse_sequence(sequence: str) -> str:
    """
    Reverses the sequence, while preserving the position of any modifications.

    Args:
        sequence (str): The amino acid sequence to be reversed.

    Returns:
        str: The reversed sequence with modifications preserved.
    """
    mods = parse_modified_sequence(sequence)
    stripped_sequence = strip_modifications(sequence)[::-1]
    mod_reverse = {len(stripped_sequence) - 1 - k if k != -1 else -1: v for k, v in mods.items()}
    return create_modified_sequence(stripped_sequence, mod_reverse)


def shift_sequence_left(sequence: str, shift: int) -> str:
    """
    Shifts the sequence to the left by a given number of positions, while preserving the position of any modifications.

    Args:
        sequence (str): The sequence to be shifted.
        shift (int): The number of positions to shift the sequence to the left.

    Returns:
        str: The shifted sequence with modifications preserved.
    """
    mods = parse_modified_sequence(sequence)
    stripped_sequence = strip_modifications(sequence)
    seq_len = len(stripped_sequence)

    # Take modulus to ensure proper wrapping around
    effective_shift = shift % seq_len

    shifted_stripped_sequence = stripped_sequence[effective_shift:] + stripped_sequence[:effective_shift]

    # Shift mod positions, taking care to avoid negative keys
    shifted_sequence_mods = {(k - effective_shift) % seq_len if k != -1 else -1: v for k, v in mods.items()}

    return create_modified_sequence(shifted_stripped_sequence, shifted_sequence_mods)


def identify_cleavage_sites(protein_sequence: str, enzyme_regex: str) -> List[int]:
    """
    Identifies cleavage sites in a protein sequence, based on enzyme_regex pattern.
    If enzyme_regex is a key in PROTEASES, the corresponding regex pattern will be used.
    Cleavage sites are identified as the positions after matching residues.

    Args:
        protein_sequence (str): The protein sequence to be processed.
        enzyme_regex (str): Regular expression pattern or key in PROTEASES dictionary that defines the cleavage
                            rule of the enzyme.

    Returns:
        List[int]: A list of positions (1-indexed) in the protein sequence where cleavage occurs.
    """

    if enzyme_regex in PROTEASES:
        enzyme_regex = PROTEASES[enzyme_regex]

    enzyme_sites = []
    for site in reg.finditer(enzyme_regex, protein_sequence, overlapped=True):
        enzyme_sites.append(site.span(0))
    return [site[0] + 1 for site in enzyme_sites]


def digest_sequence(sequence: str, enzyme_regex: Union[List[str], str], missed_cleavages: int, min_len: int = None,
                    max_len: int = None, semi: bool = False) -> List[
    str]:
    """
    Digests a given amino acid sequence using specified enzyme rules and parameters, returning a list of peptides.

    The sequence is broken down into peptides based on the enzyme's cleavage rules, with the number of
    missed cleavages taken into account. The returned peptides are filtered based on minimum and maximum length.

    Args:
        sequence (str): The amino acid sequence to be digested.
        enzyme_regex (str): Regular expression representing the enzyme's cleavage rules.
        missed_cleavages (int): The number of allowed missed cleavages.
        min_len (int): Minimum length of the returned peptides.
        max_len (int): Maximum length of the returned peptides.
        semi (bool): Whether to include semi-digested peptides.

    Returns:
        List[str]: The list of digested peptides.
    """

    if isinstance(enzyme_regex, str):
        enzyme_regex = [enzyme_regex]

    cleavage_sites = []
    for regex in enzyme_regex:
        cleavage_sites.extend(identify_cleavage_sites(sequence, regex))
    cleavage_sites.sort()

    spans = build_spans(len(sequence), cleavage_sites, missed_cleavages, min_len, max_len, semi)
    sequences = [span_to_sequence(sequence, span) for span in spans]

    return sequences


def get_fragment_sequences(sequence: str, ion_type: str) -> List[str]:
    """
    Generate fragment sequences for a given peptide sequence based on the specified ion type

    Args:
        sequence (str): The peptide sequence.
        ion_type (str): The type of ion for which fragments are generated.

    Returns:
        List[str]: The fragment sequences.
    """

    return list(sequence_generator(sequence, forward=ion_type in 'xyz'))


def get_internal_fragment_sequences(sequence: str, ion_type:str) -> List[str]:
    """
    Generate internal fragment sequences for a given peptide sequence based on the specified ion type.

    Args:
        sequence (str): The peptide sequence.
        ion_type (str): The type of ion for which internal fragments are generated.

    Returns:
        List[str]: The internal fragment sequences.
    """
    return list(sequence_generator(sequence, forward=ion_type not in 'xyz'))


import unittest

from peptacular.spans import build_non_enzymatic_spans, build_left_semi_spans, build_right_semi_spans, span_to_sequence, \
    get_enzymatic_spans, get_semi_spans


class TestSpans(unittest.TestCase):

    def test_get_non_enzymatic_spans(self):
        span = (0, 4, 1)
        expected_output = [(0, 1, 0), (0, 2, 0), (0, 3, 0), (0, 4, 0), (1, 2, 0), (1, 3, 0), (1, 4, 0), (2, 3, 0),
                           (2, 4, 0), (3, 4, 0)]
        self.assertEqual(build_non_enzymatic_spans(span), expected_output)

        span = (0, 3, 2)
        min_len = 1
        max_len = 2
        expected_output = [(0, 1, 0), (0, 2, 0), (1, 2, 0), (1, 3, 0), (2, 3, 0)]
        self.assertEqual(build_non_enzymatic_spans(span, min_len, max_len), expected_output)

    def test_get_left_spans(self):
        span = (0, 4, 1)
        expected_output = [(0, 3, 1), (0, 2, 1), (0, 1, 1)]
        self.assertEqual(build_left_semi_spans(span), expected_output)

        span = (0, 3, 2)
        min_len = 1
        max_len = 2
        expected_output = [(0, 2, 2), (0, 1, 2)]
        self.assertEqual(build_left_semi_spans(span, min_len, max_len), expected_output)

    def test_get_right_spans(self):
        span = (0, 4, 1)
        expected_output = [(1, 4, 1), (2, 4, 1), (3, 4, 1)]
        self.assertEqual(build_right_semi_spans(span), expected_output)

        span = (0, 3, 2)
        min_len = 1
        max_len = 2
        expected_output = [(1, 3, 2), (2, 3, 2)]
        self.assertEqual(build_right_semi_spans(span, min_len, max_len), expected_output)

    def test_span_to_sequence(self):
        sequence = "ABCDEFGH"
        span = (0, 4, 1)
        expected_output = "ABCD"
        self.assertEqual(span_to_sequence(sequence, span), expected_output)

        sequence = "IJKLMNOP"
        span = (2, 6, 1)
        expected_output = "KLMN"
        self.assertEqual(span_to_sequence(sequence, span), expected_output)

    def test_get_non_enzymatic_spans_no_sub_spans(self):
        span = (0, 0, 1)
        expected_output = []
        self.assertEqual(build_non_enzymatic_spans(span), expected_output)

    def test_get_left_spans_span_of_one(self):
        span = (0, 0, 1)
        expected_output = []
        self.assertEqual(build_left_semi_spans(span), expected_output)

    def test_get_right_spans_span_of_one(self):
        span = (0, 0, 1)
        expected_output = []
        self.assertEqual(build_right_semi_spans(span), expected_output)

    def test_span_to_sequence_empty_sequence(self):
        sequence = ""
        span = (0, 0, 1)
        expected_output = ""
        self.assertEqual(span_to_sequence(sequence, span), expected_output)

    def test_span_to_sequence_span_beyond_sequence_length(self):
        sequence = "ABC"
        span = (0, 10, 1)
        expected_output = "ABC"
        self.assertEqual(span_to_sequence(sequence, span), expected_output)

    def test_span_to_sequence_negative_span(self):
        sequence = "ABCDEFGH"
        span = (-4, -1, 1)
        with self.assertRaises(ValueError):
            span_to_sequence(sequence, span)

    def test_get_non_enzymatic_spans_non_integer_input(self):
        span = (0.5, 4.5, 1)
        with self.assertRaises(TypeError):
            build_non_enzymatic_spans(span)

    def test_get_non_enzymatic_spans_invalid_span(self):
        span = (4, 0, 1)
        with self.assertRaises(ValueError):
            build_non_enzymatic_spans(span)

    def test_get_left_spans_non_integer_input(self):
        span = (0.5, 4.5, 1)
        with self.assertRaises(TypeError):
            build_left_semi_spans(span)

    def test_get_left_spans_invalid_span(self):
        span = (4, 0, 1)
        with self.assertRaises(ValueError):
            build_left_semi_spans(span)

    def test_get_right_spans_non_integer_input(self):
        span = (0.5, 4.5, 1)
        with self.assertRaises(TypeError):
            build_right_semi_spans(span)

    def test_get_right_spans_invalid_span(self):
        span = (4, 0, 1)
        with self.assertRaises(ValueError):
            build_right_semi_spans(span)

    def test_span_to_sequence_non_integer_input(self):
        sequence = "ABCDEFGH"
        span = (0.5, 4.5, 1)
        with self.assertRaises(TypeError):
            span_to_sequence(sequence, span)

    def test_span_to_sequence_invalid_span(self):
        sequence = "ABCDEFGH"
        span = (4, 0, 1)
        with self.assertRaises(ValueError):
            span_to_sequence(sequence, span)

    def test_get_enzymatic_spans(self):
        end_index = 10
        enzyme_sites = [2, 5, 8]
        missed_cleavages = 1
        min_len = 2
        max_len = 6

        expected_spans = [(0, 2, 0), (0, 5, 1), (2, 5, 0), (2, 8, 1), (5, 8, 0), (5, 10, 1), (8, 10, 0)]
        spans = get_enzymatic_spans(end_index, enzyme_sites, missed_cleavages, min_len, max_len)
        self.assertEqual(spans, expected_spans)

    def test_get_enzymatic_spans2(self):
        end_index = 10
        enzyme_sites = [2, 5, 8]
        missed_cleavages = 1
        min_len = 2
        max_len = 5

        expected_spans = [(0, 2, 0), (0, 5, 1), (2, 5, 0), (5, 8, 0), (5, 10, 1), (8, 10, 0)]
        spans = get_enzymatic_spans(end_index, enzyme_sites, missed_cleavages, min_len, max_len)
        self.assertEqual(spans, expected_spans)

    def test_get_enzymatic_spans_oob(self):
        end_index = 10
        enzyme_sites = [2, 5, 8]
        missed_cleavages = 3
        min_len = None
        max_len = None

        expected_spans = [(0, 2, 0), (0, 5, 1), (0, 8, 2), (0, 10, 3), (2, 5, 0), (2, 8, 1), (2, 10, 2), (5, 8, 0),
                          (5, 10, 1), (8, 10, 0)]
        spans = get_enzymatic_spans(end_index, enzyme_sites, missed_cleavages, min_len, max_len)
        self.assertEqual(spans, expected_spans)

    def test_get_semi_spans(self):
        spans = [(0, 5, 0), (0, 8, 1)]
        min_len = 2
        max_len = 5

        expected_semi_spans = [(0, 4, 0), (0, 3, 0), (0, 2, 0), (1, 5, 0), (2, 5, 0), (3, 5, 0), (3, 8, 1),
                               (4, 8, 1), (5, 8, 1), (6, 8, 1)]
        semi_spans = get_semi_spans(spans, min_len, max_len)
        self.assertEqual(semi_spans, expected_semi_spans)

    def test_get_semi_spans2(self):
        spans = [(10, 11, 0), (10, 15, 1), (10, 18, 2)]
        min_len = 2
        max_len = 5

        expected_semi_spans = [(10, 14, 1), (10, 13, 1), (10, 12, 1), (11, 15, 1), (12, 15, 1), (13, 15, 1),
                               (13, 18, 2),
                               (14, 18, 2), (15, 18, 2), (16, 18, 2)]
        semi_spans = get_semi_spans(spans, min_len, max_len)
        self.assertEqual(semi_spans, expected_semi_spans)

    def test_get_semi_spans3(self):
        spans = [(5, 15, 2), (10, 15, 1), (14, 15, 0)]
        min_len = 2
        max_len = 5

        expected_semi_spans = [(5, 10, 2), (5, 9, 2), (5, 8, 2), (5, 7, 2), (10, 14, 1), (10, 13, 1), (10, 12, 1),
                               (11, 15, 1), (12, 15, 1), (13, 15, 1)]
        semi_spans = get_semi_spans(spans, min_len, max_len)
        self.assertEqual(semi_spans, expected_semi_spans)


if __name__ == "__main__":
    unittest.main()



from dataclasses import dataclass
from typing import List

from peptacular.mass import calculate_mz
from peptacular.sequence import get_fragment_sequences, get_internal_fragment_sequences, strip_modifications


@dataclass
class Fragment:
    """
    Represents a fragment resulting from breaking a sequence at specific points.

    Attributes:
        sequence (str): The sequence of the fragment.
        mass (float): The mass-to-charge (m/z) ratio of the fragment.
        charge (int): The charge state of the fragment.
        ion_type (str): The type of ion formed by the fragmentation.
        number (int): The fragment number (starting from the C-terminus).
        internal (bool): True if the fragment is an internal fragment, False if it is a terminal fragment.
        parent_number (int): The number of the parent fragment from which the internal fragment is derived.
    """

    sequence: str
    mass: float
    charge: int
    ion_type: str
    number: int
    internal: bool
    parent_number: int

    @property
    def start(self):
        if self.ion_type in 'abc':
            if self.internal is False:
                return 0
            else:
                return self.parent_number - self.number
        elif self.ion_type in 'xyz':
            return -self.parent_number

    @property
    def end(self):
        if self.ion_type in 'abc':
            return self.parent_number
        elif self.ion_type in 'xyz':
            if self.internal is False:
                return None
            else:
                return self.start + self.number

    @property
    def label(self):
        return f"{'+' * self.charge}" \
               f"{self.ion_type}" \
               f"{self.parent_number}" \
               f"{'i' if self.internal else ''}" \
               f"{self.number if self.internal else ''}"


def build_fragments(sequence: str, ion_types: List[str], charges: List[int], monoisotopic: bool,
                    internal_fragments: bool) -> List[Fragment]:
    """
    Generate a list of fragments for a given peptide sequence.

    Args:
        sequence (str): The peptide sequence (can be modified).
        ion_types (List[str]): A list of ion types to consider, e.g., ['b', 'y'].
        charges (List[int]): A list of charge states for the fragments.
        monoisotopic (bool): If True, use monoisotopic masses for calculations.
        internal_fragments (bool): If True, include internal fragments; otherwise, only terminal fragments are generated.

    Returns:
        List[Fragment]: A list of Fragment objects representing the generated fragments.

    Example:
        sequence = "PEPTIDE"
        ion_types = ['b', 'y']
        charges = [1, 2]
        monoisotopic = True
        internal_fragments = True

        fragments = build_fragments(sequence, ion_types, charges, monoisotopic, internal_fragments)
        for fragment in fragments:
            print(fragment.sequence, fragment.mass, fragment.charge, fragment.ion_type, fragment.number,
                  fragment.internal, fragment.parent_number)
    """
    fragments = []
    for ion_type in ion_types:
        seq_length = len(strip_modifications(sequence))
        for number, fragment_sequence in enumerate(get_fragment_sequences(sequence=sequence, ion_type=ion_type)):

            frag_number = seq_length - number
            for charge in charges:
                mass = calculate_mz(fragment_sequence, charge=charge, ion_type=ion_type, monoisotopic=monoisotopic)
                fragments.append(Fragment(sequence=fragment_sequence, mass=mass, charge=charge, ion_type=ion_type,
                                          number=frag_number, internal=False, parent_number=frag_number))
            if internal_fragments is True:
                fragment_seq_len = len(strip_modifications(fragment_sequence))
                for internal_number, internal_fragment_sequence in enumerate(
                        get_internal_fragment_sequences(sequence=fragment_sequence, ion_type=ion_type)):

                    if internal_number == 0:  # skip first since its already created
                        continue

                    internal_frag_number = fragment_seq_len - internal_number

                    for charge in charges:
                        mass = calculate_mz(internal_fragment_sequence, charge=charge, ion_type=ion_type,
                                            monoisotopic=monoisotopic)
                        fragments.append(
                            Fragment(sequence=internal_fragment_sequence, mass=mass, charge=charge, ion_type=ion_type,
                                     number=internal_frag_number, internal=True, parent_number=frag_number))

    return fragments
